#include <stdio.h>

#  define CUDA_SAFE_CALL( call) do {                                         \
    cudaError err = call;                                                    \
    if( cudaSuccess != err) {                                                \
        fprintf(stderr, "Cuda error in file '%s' in line %i : %s.\n",        \
                __FILE__, __LINE__, cudaGetErrorString( err) );              \
    exit(EXIT_FAILURE);                                                      \
    } } while (0)

#ifdef NDEBUG
#define CUT_CHECK_ERROR(errorMessage)
#else
 #  define CUT_CHECK_ERROR(errorMessage) do {                                 \
    cudaThreadSynchronize();                                                \
    cudaError_t err = cudaGetLastError();                                    \
    if( cudaSuccess != err) {                                                \
        fprintf(stderr, "Cuda error: %s in file '%s' in line %i : %s.\n",    \
                errorMessage, __FILE__, __LINE__, cudaGetErrorString( err) );\
        exit(EXIT_FAILURE);                                                  \
    } } while (0)
#endif

#define BLOCK_SIZE 128

typedef __align__(16) struct
{
    char a,b,c,d;
    char e,f,g,h;
    char i,j,k,l;
    char m,n,o,p;
} t1;

typedef __align__(16) struct
{
    char a,b,c,d;
    int e;
    int f;
    int g;
} t2;

typedef __align__(16) struct
{
    char a,b;
    short c;
    int d;
    int e;
} t3;

typedef __align__(16) struct
{
    short a,b;
    int c;
    int d;
    int e;
} t4;

typedef __align__(16) struct
{
    short a,b;
    int c;
    int d;
} t5;

typedef __align__(16) struct
{
    int a;
    int b;
    int c;
    int d;
} t6;

t1 t1var = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
t2 t2var = {0,0,0,0,0,0,0};
t3 t3var = {0,0,0,0,0};
t4 t4var = {0,0,0,0,0};
t5 t5var = {0,0,0,0};
t6 t6var = {0,0,0,0};

template <class T> __global__ void copy_gmem(T* g_idata, T* g_odata, T c)
{
    const unsigned int idx = threadIdx.x + blockIdx.x * blockDim.x;
    g_odata[idx] = g_idata[idx];
}

template <class T> __global__ void write_only(T* g_idata, T* g_odata, T c)
{
    const unsigned int idx = threadIdx.x + blockIdx.x * blockDim.x;
    g_odata[idx] = c;
}

template <class T> __global__ void read_only_gmem(T* g_idata, T* g_odata, T c)
{
    const unsigned int idx = threadIdx.x + blockIdx.x * blockDim.x;
    __shared__ T shared[BLOCK_SIZE];
    shared[threadIdx.x] = g_idata[idx];
    *((float *)(&shared[(threadIdx.x + 1) & (BLOCK_SIZE-1)])) += 1.0;
}


#define DO_BMARK(kernel, type, value, name, ntransfer)                  \
{                                                                       \
    kernel<<< grid, threads >>>((type *)d_idata, (type *)d_odata, value); \
                                                                        \
    cudaEvent_t start, end;                                             \
    CUDA_SAFE_CALL( cudaEventCreate(&start) );                          \
    CUDA_SAFE_CALL( cudaEventCreate(&end) );                            \
                                                                        \
    CUDA_SAFE_CALL( cudaEventRecord(start, 0) );                        \
    for (int i=0; i < nIters; ++i)                                      \
        kernel<<< grid, threads >>>((type *)d_idata, (type *)d_odata, value); \
                                                                        \
    CUDA_SAFE_CALL( cudaEventRecord(end, 0) );                          \
    CUDA_SAFE_CALL( cudaEventSynchronize(end) );                        \
                                                                        \
    float runTime;                                                      \
    CUDA_SAFE_CALL( cudaEventElapsedTime(&runTime, start, end) );       \
    runTime /= float(nIters);                                           \
    printf("%s - Bandwidth:    %f GiB/s\n", name, (ntransfer * len * sizeof(type)) / (runTime * 1.0e-3 * 1024*1024*1024)); \
    CUDA_SAFE_CALL( cudaEventDestroy(start) );                          \
    CUDA_SAFE_CALL( cudaEventDestroy(end) );                            \
}

int main()
{
    int len = 1 << 22;
    int num_threads = BLOCK_SIZE;
    int nIters = 1;
    
    float4 *d_idata, *d_odata;
    CUDA_SAFE_CALL( cudaMalloc((void**)&d_idata, sizeof(float4)*len) );
    CUDA_SAFE_CALL( cudaMalloc((void**)&d_odata, sizeof(float4)*len) );
    
    dim3  threads(num_threads, 1, 1);
    dim3  grid(len/num_threads, 1, 1);
    
    DO_BMARK(copy_gmem<float>, float, 0.0f, "copy_gmem<float>", 2);
    DO_BMARK(copy_gmem<float2>, float2, make_float2(0.0f, 0.0f), "copy_gmem<float2>", 2);
    DO_BMARK(copy_gmem<float4>, float4, make_float4(0.0f, 0.0f, 0.0f, 0.0f), "copy_gmem<float4>", 2);
    DO_BMARK(copy_gmem<t1>, t1, t1var, "copy_gmem<t1>", 2);
    DO_BMARK(copy_gmem<t2>, t2, t2var, "copy_gmem<t2>", 2);
    DO_BMARK(copy_gmem<t3>, t3, t3var, "copy_gmem<t3>", 2);
    DO_BMARK(copy_gmem<t4>, t4, t4var, "copy_gmem<t4>", 2);
    DO_BMARK(copy_gmem<t5>, t5, t5var, "copy_gmem<t5>", 2);
    DO_BMARK(copy_gmem<t6>, t6, t6var, "copy_gmem<t6>", 2);
    
    printf("\n");
    DO_BMARK(write_only<float>, float, 0.0f, "write_only<float>", 1);
    DO_BMARK(write_only<float2>, float2, make_float2(0.0f, 0.0f), "write_only<float2>", 1);
    DO_BMARK(write_only<float4>, float4, make_float4(0.0f, 0.0f, 0.0f, 0.0f), "write_only<float4>", 1);
    DO_BMARK(write_only<t1>, t1, t1var, "write_only<t1>", 1);
    DO_BMARK(write_only<t2>, t2, t2var, "write_only<t2>", 1);
    DO_BMARK(write_only<t3>, t3, t3var, "write_only<t3>", 1);
    DO_BMARK(write_only<t4>, t4, t4var, "write_only<t4>", 1);
    DO_BMARK(write_only<t5>, t5, t5var, "write_only<t5>", 1);
    DO_BMARK(write_only<t6>, t6, t6var, "write_only<t6>", 1);
    
    printf("\n");
    DO_BMARK(read_only_gmem<float>, float, 0.0f, "read_only_gmem<float>", 1);
    DO_BMARK(read_only_gmem<float2>, float2, make_float2(0.0f, 0.0f), "read_only_gmem<float2>", 1);
    DO_BMARK(read_only_gmem<float4>, float4, make_float4(0.0f, 0.0f, 0.0f, 0.0f), "read_only_gmem<float4>", 1);
    DO_BMARK(read_only_gmem<t1>, t1, t1var, "read_only_gmem<t1>", 1);
    DO_BMARK(read_only_gmem<t2>, t2, t2var, "read_only_gmem<t2>", 1);
    DO_BMARK(read_only_gmem<t3>, t3, t3var, "read_only_gmem<t3>", 1);
    DO_BMARK(read_only_gmem<t4>, t4, t4var, "read_only_gmem<t4>", 1);
    DO_BMARK(read_only_gmem<t5>, t5, t5var, "read_only_gmem<t5>", 1);
    DO_BMARK(read_only_gmem<t6>, t6, t6var, "read_only_gmem<t6>", 1);
}
